var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { basename, dirname, join, resolve } from 'node:path';
import { asArray, existsSync, globby, read, readJson, remove, resolveTsConfig, unixify, write, } from './util.mjs';
export const DEFAULT_FIX_OPTIONS = {
    cwd: process.cwd(),
    tsconfig: './tsconfig.json',
    filenameVar: true,
    dirnameVar: true,
    ext: true,
    unlink: true,
    debug: () => { }, // eslint-disable-line
};
export const normalizeOptions = (opts) => (Object.assign(Object.assign(Object.assign({}, DEFAULT_FIX_OPTIONS), opts), { debug: typeof (opts === null || opts === void 0 ? void 0 : opts.debug) === 'function'
        ? opts.debug
        : (opts === null || opts === void 0 ? void 0 : opts.debug) === true
            ? console.log
            : DEFAULT_FIX_OPTIONS.debug }));
export const findTargets = (tsconfig, cwd) => asArray(tsconfig).reduce((targets, file) => {
    var _a, _b, _c, _d;
    const tsconfigJson = resolveTsConfig(resolve(cwd, file));
    const outDir = (_a = tsconfigJson === null || tsconfigJson === void 0 ? void 0 : tsconfigJson.compilerOptions) === null || _a === void 0 ? void 0 : _a.outDir;
    const module = (_d = (_b = tsconfigJson === null || tsconfigJson === void 0 ? void 0 : tsconfigJson.compilerOptions) === null || _b === void 0 ? void 0 : (_c = _b.module).toLowerCase) === null || _d === void 0 ? void 0 : _d.call(_c);
    if (outDir && module.startsWith('es')) {
        targets.push(outDir);
    }
    else {
        console.warn('tsconfig should declare `outDir` and `module` type es6 or above');
    }
    return targets;
}, []);
export const resolveDependency = (parent, nested, files, cwd) => {
    const dir = dirname(parent);
    const nmdir = resolve(cwd, 'node_modules');
    const bases = /^\..+\.[^./\\]+$/.test(nested)
        ? [nested, nested.replace(/\.[^./\\]+$/, '')]
        : [nested];
    const variants = ['.js', '.cjs', '.mjs'].reduce((m, e) => {
        bases.forEach((v) => m.push(`${v}${e}`, `${v}/index${e}`));
        return m;
    }, []);
    return (variants.find((f) => files.includes(unixify(resolve(dir, f)))) ||
        variants.find((f) => files.includes(unixify(resolve(nmdir, f)))) ||
        nested);
};
export const fixFilenameExtensions = (names, ext) => names.map((name) => name.endsWith('.d.ts')
    ? name
    : name.replace(/\.[^./\\]+$/, ext));
export const fixModuleReferences = (contents, filename, filenames, cwd, ignore) => contents.replace(/((?:\s|^)import\s+|\s+from\s+|\W(?:import|require)\s*\()(["'])([^"']+\/[^"']+|\.{1,2})\/?(["'])/g, (_matched, control, q1, from, q2) => `${control}${q1}${ignore.includes(from) ? from : resolveDependency(filename, from, filenames, cwd)}${q2}`);
export const fixDirnameVar = (contents, isSource) => contents.replace(/__dirname/g, `\`\${process.platform === 'win32' ? '' : '/'}\${/file:\\/{2,3}(.+)\\/[^/]/.exec(import.meta.url)${isSource ? '!' : ''}[1]}\``); // eslint-disable-line
export const fixFilenameVar = (contents, isSource) => contents.replace(/__filename/g, `\`\${process.platform === 'win32' ? '' : '/'}\${/file:\\/{2,3}(.+)/.exec(import.meta.url)${isSource ? '!' : ''}[1]}\``); // eslint-disable-line
export const fixDefaultExport = (contents) => contents.includes('export default')
    ? contents
    : `${contents}
export default undefined
`;
export const fixBlankFiles = (contents) => contents.trim().length === 0
    ? `
export {}
export default undefined
` : contents;
export const fixSourceMapRef = (contents, originName, filename) => originName === filename
    ? contents
    : contents.replace(`//# sourceMappingURL=${basename(originName)}.map`, `//# sourceMappingURL=${basename(filename)}.map`);
export const fixContents = (contents, filename, filenames, { cwd, ext, dirnameVar, filenameVar, fillBlank, forceDefaultExport, sourceMap }, originName = filename, // NOTE Weird contract to avoid breaking change
isSource = false, ignore = []) => {
    let _contents = contents;
    if (ext) {
        _contents = fixModuleReferences(_contents, filename, filenames, cwd, ignore);
    }
    if (dirnameVar) {
        _contents = fixDirnameVar(_contents, isSource);
    }
    if (filenameVar) {
        _contents = fixFilenameVar(_contents, isSource);
    }
    if (fillBlank) {
        _contents = fixBlankFiles(_contents);
    }
    if (forceDefaultExport) {
        _contents = fixDefaultExport(_contents);
    }
    if (sourceMap) {
        _contents = fixSourceMapRef(_contents, originName, filename);
    }
    return _contents;
};
const resolvePrefix = (prefix, pattern) => {
    if (!pattern) {
        return prefix;
    }
    let _pattern = pattern;
    if (_pattern.includes('*')) {
        _pattern = _pattern.slice(0, _pattern.indexOf('*'));
        if (_pattern.includes('/')) {
            _pattern = _pattern.slice(0, _pattern.lastIndexOf('/'));
        }
    }
    return join(prefix, _pattern);
};
const getExportsEntries = (exports) => {
    const entries = Object.entries(exports);
    const parseConditional = (e) => typeof e === 'string' ? [e] : Object.values(e).map(parseConditional).flat(2);
    // has subpaths
    if (typeof exports !== 'string' && Object.keys(exports).some((k) => k.startsWith('.'))) {
        return entries.map(([k, v]) => [k, parseConditional(v)]);
    }
    return [['.', parseConditional(exports)]];
};
const getExternalEsmModules = (cwd) => globby(['node_modules/*/package.json', 'node_modules/@*/*/package.json'], {
    cwd,
    onlyFiles: true,
    absolute: true,
}).then((files) => __awaiter(void 0, void 0, void 0, function* () {
    return (yield Promise.all(files
        .map((f) => __awaiter(void 0, void 0, void 0, function* () {
        const { name, exports } = yield readJson(f);
        if (!exports) {
            return { name };
        }
        const _dir = dirname(f);
        const exportsEntries = getExportsEntries(exports);
        return {
            name,
            files: (yield Promise.all(exportsEntries.map(([key, values]) => Promise.all(values.map((value) => __awaiter(void 0, void 0, void 0, function* () {
                return (yield globby(value, { cwd: _dir, onlyFiles: true, absolute: false }))
                    .map(file => join(file)
                    .replace(resolvePrefix('.', value), resolvePrefix(name, key)));
            })))))).flat(2)
        };
    })))).reduce((m, { name, files: _files }) => {
        if (name) {
            m.names.push(name);
        }
        if (_files) {
            m.files.push(..._files);
        }
        return m;
    }, { names: [], files: [] });
}));
const getExternalModules = (cwd) => __awaiter(void 0, void 0, void 0, function* () {
    const { names, files: esmModules } = yield getExternalEsmModules(cwd);
    const cjsModules = yield globby([
        '!node_modules/.cache',
        '!node_modules/.bin',
        '!node_modules/**/node_modules',
        ...names.map(m => `!node_modules/${m}`),
        'node_modules/**/*.(m|c)?js',
    ], {
        cwd,
        onlyFiles: true,
        absolute: true,
    });
    return {
        cjsModules,
        esmModules,
        allPackages: names,
    };
});
export const getPatterns = (sources, targets) => sources.length > 0
    ? sources.map((src) => src.includes('*') ? src : `${src}/**/*.{ts,tsx}`)
    : targets.map((target) => target.includes('*') ? target : `${target}/**/*.{js,d.ts}`);
export const fix = (opts) => __awaiter(void 0, void 0, void 0, function* () {
    const _opts = normalizeOptions(opts);
    const { cwd, target, src, tsconfig, out = cwd, ext, debug, unlink, sourceMap } = _opts;
    const outDir = resolve(cwd, out);
    const sources = asArray(src);
    const targets = [...asArray(target), ...findTargets(tsconfig, cwd)];
    debug('debug:cwd', cwd);
    debug('debug:outdir', outDir);
    debug('debug:sources', sources);
    debug('debug:targets', targets);
    const isSource = sources.length > 0;
    const patterns = getPatterns(sources, targets);
    const localModules = yield globby(patterns, {
        cwd,
        onlyFiles: true,
        absolute: true,
    });
    const { cjsModules, esmModules, allPackages } = yield getExternalModules(cwd);
    debug('debug:external-cjs-modules', cjsModules);
    debug('debug:external-esm-modules', esmModules);
    const ignore = [...esmModules, ...allPackages];
    const _localModules = typeof ext === 'string' ? fixFilenameExtensions(localModules, ext) : localModules;
    const allModules = [...cjsModules, ..._localModules];
    const allJsModules = [...cjsModules, ...fixFilenameExtensions(localModules, '.js')];
    debug('debug:local-modules', _localModules);
    _localModules.forEach((name, i) => {
        // NOTE d.ts may refer to .js ext only
        const all = name.endsWith('.d.ts') ? allJsModules : allModules;
        const originName = localModules[i];
        const nextName = (sources.length === 0 ? name : originName).replace(unixify(cwd), unixify(outDir));
        const contents = read(originName);
        const _contents = fixContents(contents, name, all, _opts, originName, isSource, ignore);
        write(nextName, _contents);
        if (sources.length === 0 && unlink && cwd === outDir && nextName !== originName) {
            remove(originName);
        }
        if (sourceMap) {
            patchSourceFile(originName, nextName, unlink && cwd === outDir);
        }
    });
});
const patchSourceFile = (name, nextName, unlink = false) => {
    if (name === nextName) {
        return;
    }
    const mapfile = `${name}.map`;
    if (!existsSync(mapfile)) {
        return;
    }
    const nextMapfile = `${nextName}.map`;
    const contents = readJson(mapfile);
    contents.file = basename(nextName);
    write(nextMapfile, JSON.stringify(contents));
    if (unlink) {
        remove(mapfile);
    }
};
